from pathlib import Path

from pycrunch_trace.filters import CustomFileFilter
from pycrunch_trace.oop import Clock, File
from pycrunch_trace.tracing.simulation import models
from pycrunch_trace.tracing.simple_tracer import SimpleTracer
from pycrunch_trace.tracing.simulator_sink import SimulationEvent


def create_event_1():
    code_clone = models.Code()
    code_clone.co_name = 'a___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 23
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_2():
    code_clone = models.Code()
    code_clone.co_name = 'a___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 24
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_3():
    code_clone = models.Code()
    code_clone.co_name = 'a___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 25
    sim_frame.f_locals = {
        'a': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_4():
    code_clone = models.Code()
    code_clone.co_name = 'a___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 26
    sim_frame.f_locals = {
        'a': 1,
        'b': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_5():
    code_clone = models.Code()
    code_clone.co_name = 'b___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 15
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_6():
    code_clone = models.Code()
    code_clone.co_name = 'b___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 16
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_7():
    code_clone = models.Code()
    code_clone.co_name = 'b___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 17
    sim_frame.f_locals = {
        'a': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_8():
    code_clone = models.Code()
    code_clone.co_name = 'c___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 8
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_9():
    code_clone = models.Code()
    code_clone.co_name = 'c___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 9
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_10():
    code_clone = models.Code()
    code_clone.co_name = 'c___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 10
    sim_frame.f_locals = {
        'a': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_11():
    code_clone = models.Code()
    code_clone.co_name = 'c___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 11
    sim_frame.f_locals = {
        'a': 1,
        'b': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_12():
    code_clone = models.Code()
    code_clone.co_name = 'd___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 3
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_13():
    code_clone = models.Code()
    code_clone.co_name = 'd___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 4
    sim_frame.f_locals = {}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_14():
    code_clone = models.Code()
    code_clone.co_name = 'd___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 5
    sim_frame.f_locals = {
        'a': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_15():
    code_clone = models.Code()
    code_clone.co_name = 'd___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 6
    sim_frame.f_locals = {
        'a': 1,
        'b': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_16():
    code_clone = models.Code()
    code_clone.co_name = 'd___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 6
    sim_frame.f_locals = {
        'a': 1,
        'b': 2,
        'c': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', None)

    return evt


def create_event_17():
    code_clone = models.Code()
    code_clone.co_name = 'c___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 12
    sim_frame.f_locals = {
        'a': 1,
        'b': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_18():
    code_clone = models.Code()
    code_clone.co_name = 'c___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 12
    sim_frame.f_locals = {
        'a': 1,
        'b': 2,
        'c': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', None)

    return evt


def create_event_19():
    code_clone = models.Code()
    code_clone.co_name = 'b___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 18
    sim_frame.f_locals = {
        'a': 1}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_20():
    code_clone = models.Code()
    code_clone.co_name = 'b___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 19
    sim_frame.f_locals = {
        'a': 1,
        'b': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_21():
    code_clone = models.Code()
    code_clone.co_name = 'b___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 19
    sim_frame.f_locals = {
        'a': 1,
        'b': 2,
        'c': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', None)

    return evt


def create_event_22():
    code_clone = models.Code()
    code_clone.co_name = 'a___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 27
    sim_frame.f_locals = {
        'a': 1,
        'b': 2}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def create_event_23():
    code_clone = models.Code()
    code_clone.co_name = 'a___'
    code_clone.co_filename = '/pycrunch_tracer/demo/demo_tree_v1.py'
    code_clone.co_argcount = 0

    sim_frame = models.Frame()
    sim_frame.f_lineno = 27
    sim_frame.f_locals = {
        'a': 1,
        'b': 2,
        'c': 3}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'return', None)

    return evt


def create_event_24():
    code_clone = models.Code()
    code_clone.co_name = 'stop'
    code_clone.co_filename = '/pycrunch_tracer/client/api/tracing.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 70
    sim_frame.f_locals = {
        'self': '< pycrunch_tracer.api.tracing'}

    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'call', None)

    return evt


def create_event_25():
    code_clone = models.Code()
    code_clone.co_name = 'stop'
    code_clone.co_filename = '/pycrunch_tracer/client/api/tracing.py'
    code_clone.co_argcount = 1

    sim_frame = models.Frame()
    sim_frame.f_lineno = 71
    sim_frame.f_locals = {   'self': '< pycrunch_tracer.api.tracing'}
    sim_frame.f_code = code_clone
    evt = SimulationEvent(sim_frame, 'line', None)

    return evt


def test_simulated():
    events = []
    events.append(create_event_1())
    events.append(create_event_2())
    events.append(create_event_3())
    events.append(create_event_4())
    events.append(create_event_5())
    events.append(create_event_6())
    events.append(create_event_7())
    events.append(create_event_8())
    events.append(create_event_9())
    events.append(create_event_10())
    events.append(create_event_11())
    events.append(create_event_12())
    events.append(create_event_13())
    events.append(create_event_14())
    events.append(create_event_15())
    events.append(create_event_16())
    events.append(create_event_17())
    events.append(create_event_18())
    events.append(create_event_19())
    events.append(create_event_20())
    events.append(create_event_21())
    events.append(create_event_22())
    events.append(create_event_23())
    events.append(create_event_24())
    events.append(create_event_25())
    package_directory = Path(__file__).parent.parent.parent
    profile_name = 'default.profile.yaml'

    f_filter = CustomFileFilter(File(package_directory.joinpath('pycrunch-profiles', profile_name)))

    sut = SimpleTracer(events, 'sim_round2', f_filter, Clock(), self.outgoingQueue)
    for (i, x) in enumerate(events):
        if i > 24:
            break
        print(i)
        sut.simple_tracer(x.frame, x.event, x.arg)




from random import Random

from . import AbstractFileFilter


class DefaultFileFilter(AbstractFileFilter):
    def should_trace(self, filename: str) -> bool:
        # start or end with
        exclusions = (
            '/Users/gleb/code/pycrunch_tracing/',
            '/Users/gleb/code/bc/briteapps-admin/',
            'module_a.py',
            'module_b.py',
            'module_c.py',
            'invalid_picker_with_exception.py',
            '/Library/Frameworks/Python.framework/Versions/3.6/lib/python3.6'
            '/Users/gleb/venv/PyCrunch/lib/python3.6/site-packages/',
            # 'copyreg.py',
            '/Users/gleb/venv/PyCrunch/lib/python3.6/site-packages/py/',
            'api/tracing.py'
        )

        # r= Random()
        # if r.choice(range(0, 10)) == 1:
        #     return True
        # return True
        if filename.endswith(exclusions) or filename.startswith(exclusions):
            return False

        return True

        if filename.endswith(end_patterns):
            return True
        if filename.startswith(start_patterns):
            return True

        print('should_trace - false', filename)
        return False


from typing import Dict, Any


class FileMap:
    # this is for size reduction
    # filename.py -> id
    files: Dict[str, int]

    def __init__(self):
        self.files = dict()
        pass

    def file_id(self, filename: str) -> int:
        file_id = self.files.get(filename)
        if not file_id:
            file_id = len(self.files) + 1
            self.files[filename] = file_id

        return file_id

    def filename_by_id(self, search_for_file_id : int):
#         slow
        for (name, file_id) in self.files.items():
            if file_id == search_for_file_id:
                return name
        return f'file {search_for_file_id} not found'

    @classmethod
    def from_reverse(cls, files: Dict[str, int]):
        x = FileMap()
        x.files = files
        return x



import sys
import uuid
from pathlib import Path

from pycrunch_trace.client.command_buffer import ArrayCommandBuffer
from pycrunch_trace.client.networking import event_queue
from pycrunch_trace.filters import CustomFileFilter
from pycrunch_trace.oop import File, Clock, SafeFilename
from pycrunch_trace.tracing.inline_profiler import inline_profiler_instance

import pyximport
pyximport.install()
from pycrunch_trace.native.native_tracer import NativeTracer

from pycrunch_trace.tracing.simple_tracer import SimpleTracer


class Trace:
    clock: Clock
    _tracer: SimpleTracer

    def __init__(self):
        self.default_host = 'http://0.0.0.0:8080'
        self.command_buffer = ArrayCommandBuffer()
        self.is_tracing = False
        self.session_name = None
        self._tracer = None
        # self._client: network_client.TracingClient = None
        self.clock = None
        self.host = None
        self.outgoingQueue = None

    def generate_session_name(self) -> str:
        return str(uuid.uuid4())

    def start(self, session_name: str = None, host: str = None, profile_name: str = None):

        if self.is_tracing:
            raise Exception('PyCrunch tracer ERROR: tracing already started')

        self.prepare_state(host, session_name)
        self.warn_if_another_tracing_set()

        # self._client = network_client.TracingClient(self.host)
        if not profile_name:
            profile_name = 'default.profile.yaml'
        package_directory = Path(__file__).parent.parent.parent
        f_filter = CustomFileFilter(File(package_directory.joinpath('pycrunch-profiles', profile_name)))
        f_filter._ensure_loaded()
        # else:
        #     f_filter = DefaultFileFilter()

        self.start_queue()

        self.clock = Clock()
        # todo maybe move command buffer to tracer?
        # self._tracer = SimpleTracer(self.command_buffer, self.session_name, f_filter, self.clock, self.outgoingQueue)
        # TODO windows test
        self._tracer = NativeTracer(session_name, self.outgoingQueue, f_filter)
        self.outgoingQueue.start()

        self.outgoingQueue.tracing_will_start(self.session_name)

        # also trace parent function
        sys._getframe().f_back.f_trace = self._tracer.simple_tracer

        sys.settrace(self._tracer.simple_tracer)

        self.is_tracing = True

    def start_queue(self):
        self.outgoingQueue = event_queue
        self.outgoingQueue.start()

    def warn_if_another_tracing_set(self):
        if sys.gettrace():
            # there is already trace
            print('PyCrunch tracer WARNING:')
            print('  -- there is already trace function set. ')
            print('  -- continuing might result in errors ')

    def prepare_state(self, host, session_name):
        if not session_name:
            self.session_name = self.generate_session_name()
        else:
            self.session_name = SafeFilename(session_name).__str__()
        if host:
            self.host = host
        else:
            self.host = self.default_host

    def stop(self):
        sys.settrace(None)

        inline_profiler_instance.print_timings()
        # import pydevd_pycharm
        # pydevd_pycharm.settrace('localhost', port=44441, stdoutToServer=True, stderrToServer=True)
        print('tracing complete, saving results')
        self.is_tracing = False
        # snapshot.save('a', self.command_buffer)
        local = False
        # local = True
        if local:
            self._tracer.session.buffer_became_available(self.command_buffer)
            self._tracer.session.save()

        self._tracer.flush_outstanding_events()
        self._tracer.finalize()

            # self._tracer.session.save()


        # self._client.push_message(self._tracer.session)
        # self._client.disconnect()



import threading
from time import sleep

import socketio
from socketio import Client

import logging

from pycrunch_trace.client.networking.commands import EventsSlice
from pycrunch_trace.client.networking.strategies.abstract_strategy import AbstractRecordingStrategy
from pycrunch_trace.events.event_buffer_in_protobuf import EventBufferInProtobuf

logger = logging.getLogger(__name__)


class OverWireRecordingStrategy(AbstractRecordingStrategy):
    sio : Client

    def __init__(self):
        self.sio = socketio.Client()
        self.host = 'http://0.0.0.0:8080'
        self.sio = None
        self.manual_reset_event = threading.Event()

        pass
    def clean(self):
        self.sio.disconnect()

    def callback(self):
        print(f'#callback : delivered')

        logger.info(f'#callback : delivered')
        print(f'#callback : event set')
        self.manual_reset_event.set()

        # logger.info(args)

    def callback_for_disconnection(self):
        print('Disconnection Callback')
        # self.sio.disconnect()

    def recording_start(self, session_id: str):
        self.sio.emit('tracing_node_event',
                      dict(
                          action='events_stream_will_start',
                          session_id=session_id),
                      )
    def recording_stop(self, session_id: str):
        sleep(1)
        self.sio.emit('tracing_node_event', dict(action='events_stream_did_complete', session_id=session_id), callback=self.callback_for_disconnection)

    def recording_slice(self, x: EventsSlice):
        x: EventsSlice = x
        # client_introspection.save_events(x.events)
        # client_introspection.print_to_console(x.files)
        events_in_payload = len(x.events)
        bytes_to_send = EventBufferInProtobuf(x.events, x.files).as_bytes()
        print(f'manual_reset_event.waiting...')
        self.manual_reset_event.wait()
        print(f' -- done waiting for manual_reset_event, sending chunk')
        payload_size = len(bytes_to_send)
        print(f' -- sending chunk {x.chunk_number} of {x.session_id} with size {payload_size}')
        self.sio.emit('tracing_node_event',
                      dict(
                          action='events_stream',
                          session_id=x.session_id,
                          event_number=x.chunk_number,
                          bytes=bytes_to_send,
                          events_in_payload=events_in_payload,
                          payload_size=payload_size),
                      callback=self.callback)
        sleep(0.01)

    def prepare(self):
        transports = ['websocket']
        # transports = ['polling']
        print('socketio connect')

        self.sio = socketio.Client()

        # self.sio.connect(url=self.host, headers=self.connection_headers() )
        self.sio.connect(url=self.host, transports=transports, headers=self.connection_headers())

        @self.sio.event
        def message(data):
            logger.info('CLIENT: I received a message!')

        @self.sio.on('my message')
        def on_message(data):
            print('on_message')
            logger.info('CLIENT: I received a message!')

        @self.sio.event
        def connect():
            print('connect')
            self.is_connected = True
            self.manual_reset_event.set()
            logger.info("CLIENT: I'm connected!")

        @self.sio.event
        def connect_error():
            print('connect_error')
            logger.info("CLIENT: The connection failed!")

        @self.sio.event
        def disconnect():
            print('disconnect')
            # put everything in garbage?
            self.is_connected = False
            logger.info("Clearing event... until connection established back")
            self.manual_reset_event.clear()

            logger.info("CLIENT: I'm disconnected!")


    def connection_headers(self):
        from pycrunch_trace.client.api.version import version
        connection_headers = dict(
            version=version,
            product='pycrunch-tracing-node',
        )
        return connection_headers




import sys
from ppretty import ppretty

def unnamed_method(x: str):
    print('unnamed_method')

def sum(a: int, b: int) -> int:
    print('!a')
    print('!b')
    unnamed_method('a')
    return a + b

def trace_calls(frame, event, arg):
    print(f'event: {event}')
    print('f_locals '+ ppretty(frame.f_locals))
    # print(ppretty(frame))
    if event != 'call':
        return
    frame__f_code = frame.f_code
    func_name = frame__f_code.co_name
    print(f'func_name: {func_name}')
    if func_name == 'write':
        # Ignore write() calls from printing
        return
    func_line_no = frame.f_lineno
    func_filename = frame__f_code.co_filename
    if not func_filename.endswith('demo.py'):
        # Ignore calls not in this module
        return
    caller = frame.f_back
    caller_line_no = caller.f_lineno
    caller_filename = caller.f_code.co_filename
    print('* Call to', func_name)
    print(f'*  on line {func_line_no} of {func_filename}')
    print(f'*  from line {caller_line_no} of {caller_filename}')
    return

print('--- begin')

sys.settrace(trace_calls)

result = sum(1, 2)
print(result)

print('--- poshel naxoi')

